# -*- coding: utf-8 -*-
import datetime as dt
import os
import pandas as pd
from time import time
import re
from enerpi.base import CONFIG, DATA_PATH, log, timeit, SENSORS
from enerpi.catalog import EnerpiCatalog


# Config:
INIT_LOG_MARK = CONFIG.get('ENERPI_SAMPLER', 'INIT_LOG_MARK', fallback='INIT ENERPI')
HDF_STORE = CONFIG.get('ENERPI_DATA', 'HDF_STORE')
HDF_STORE_PATH = os.path.join(DATA_PATH, HDF_STORE)

KEY = CONFIG.get('ENERPI_DATA', 'KEY', fallback='/rms')
CONFIG_CATALOG = dict(raw_file=HDF_STORE,
                      key_raw_data=KEY,
                      key_summary_data='/hours',
                      key_summary_extra='/days',
                      check_integrity=True,
                      archive_existent=False,
                      verbose=False)


def init_catalog(sensors=None, base_path=DATA_PATH, **kwargs):
    """
    Get ENERPI data catalog for access & operation with params.

    :param sensors: Sensors config object (class EnerpiSamplerConf)
    :param base_path: :str: ENERPIDATA base path
    :param kwargs: :dict: parameters
    :return: :EnerpiCatalog:
    """
    conf = CONFIG_CATALOG.copy()
    conf.update(base_path=base_path)
    if kwargs:
        conf.update(**kwargs)
    return EnerpiCatalog(sensors=sensors, **conf)


def _clean_store_path(path_st):
    if os.path.pathsep not in path_st:
        path_st = os.path.join(DATA_PATH, path_st)
    else:
        path_st = os.path.abspath(path_st)
    if not os.path.splitext(path_st)[1]:
        path_st += '.h5'
    return path_st


def show_info_data(df, df_consumo=None):
    """
    Prints some info about DATA (& opt SUMMARY_DATA)

    :param df: :pd.DataFrame: DATA
    :param df_consumo: :pd.DataFrame: SUMMARY_DATA
    """
    log('DATAFRAME INFO:\n* Head:\n{}'.format(df.head()), 'info', True, False)
    log('* Tail:\n{}'.format(df.tail()), 'info', True, False)
    log('* Count & types:\n{}'
        .format(pd.concat([df.count().rename('N_rows'), df.dtypes.rename('dtypes'), df.describe().drop('count').T],
                          axis=1)), 'info', True, False)
    if df_consumo is not None and not df_consumo.empty:
        log('\n** HOURLY ELECTRICITY CONSUMPTION (kWh):\n{}'.format(df_consumo), 'magenta', True, False)
        dias = df_consumo.resample('1D').sum()
        dias['t_ref'] /= 24
        log('\n*** DAILY ELECTRICITY CONSUMPTION (kWh):\n{}'.format(dias), 'ok', True, False)


def _notify_error_in_save_raw_data(msg_error):
    from time import sleep
    from enerpi.notifier import push_enerpi_error

    t = push_enerpi_error('SAVE RAW DATA', msg_error)
    sleep(1)
    return t


def save_raw_data(data=None, path_st=HDF_STORE_PATH, catalog=None, verb=True):
    """
    Used in a subprocess launched from enerpimeter, this functions appends new *raw data* to the HDF raw store,
     and, if data-catalog is not None, updates it.
    :param data:
    :param path_st:
    :param catalog:
    :param verb:
    :return:
    """
    try:
        df_tot = None
        if data is not None and type(data) is not pd.DataFrame:
            data = pd.DataFrame(data, columns=SENSORS.columns_sampling
                                ).set_index(SENSORS.ts_column).dropna().astype(float)
            mode = 'a' if os.path.exists(path_st) else 'w'
            try:
                with pd.HDFStore(path_st, mode=mode, complevel=9, complib='blosc') as st:
                    st.append(KEY, data)
                    if catalog is not None:
                        df_tot = st[KEY]
                        log('Size Store: {:.1f} KB, {} rows'.format(os.path.getsize(path_st) / 1000, len(df_tot)),
                            'debug', verb)
            except OSError as e:
                msg_error = 'OSError "{}" trying to open "{}" in "{}" mode (save_in_store)'.format(e, path_st, mode)
                log(msg_error, 'error', True)
                _notify_error_in_save_raw_data(msg_error)
                return -1
            if catalog is not None:
                try:
                    catalog.update_catalog(data=df_tot)
                except Exception as e:
                    msg_error = 'Exception "{}" [{}] en update_catalog (save_in_store)'.format(e, e.__class__)
                    log(msg_error, 'error', True)
                    _notify_error_in_save_raw_data(msg_error)
                    return -1
        return True
    except ValueError as e:
        log('ValueError en save_in_store: {}'.format(e), 'error', True)
        return -1


@timeit('get_ts_last_save')
def get_ts_last_save(path_st=HDF_STORE_PATH, get_last_sample=False, verbose=True, n=3):
    """
    Returns last data timestamp in hdf store.

    :param path_st: :str: hdf store file path
    :param get_last_sample: :bool: returns ts or pd.DataFrame
    :param verbose: :bool: shows logging msgs in stdout.
    :param n: :int: # of tail rows
    :return: pd.Timestamp or pd.DataFrame
    """
    tic = time()
    try:
        ts = dt.datetime.fromtimestamp(os.path.getmtime(path_st))
        size_kb = os.path.getsize(path_st) / 1000
        if get_last_sample:
            with pd.HDFStore(path_st, mode='r') as st:
                try:
                    df = st.select(KEY, start=-n)
                    log('Store UPDATE: {:%c} , SIZE = {:.2f} KB. TOOK {:.3f} s'.format(ts, size_kb, time() - tic),
                        'debug', verbose)
                    return df
                except KeyError:
                    log('ERROR: Data "{}" not found in store "{}"'.format(KEY, path_st), 'err', True)
                    return ts
        log('Store UPDATE: {:%c} , SIZE = {:.2f} KB. TOOK {:.3f} s'.format(ts, size_kb, time() - tic), 'debug', verbose)
        return ts
    except FileNotFoundError:
        log('ERROR: Store not found in {}'.format(path_st), 'err', True)
        return None


def delete_log_file(log_file, verbose=True):
    """
    Removes (logging) file from disk.

    :param log_file: :str: logging file path
    :param verbose: :bool: shows logging msgs in stdout.
    """
    if os.path.exists(log_file) and os.path.isfile(log_file):
        log('Deleting LOG FILE in {} ...'.format(log_file), 'warn', verbose, False)
        os.remove(log_file)
        return True
    return False


@timeit('extract_log_file')
def extract_log_file(log_file, extract_temps=True, verbose=True):
    """
    Extracts pd.DataFrame from logging file.

    :param log_file: :str: logging file path
    :param extract_temps: :bool: process RPI temperature logging entries (appends columns 'CPU' & 'GPU')
    :param verbose: :bool: shows logging msgs in stdout.
    :return: time-indexed pd.DataFrame
    """
    if os.path.exists(log_file):
        rg_log_msg = re.compile('(?P<tipo>INFO|WARNING|DEBUG|ERROR) \[(?P<func>.+?)\] '
                                '- (?P<ts>\d{1,2}/\d\d/\d\d\d\d \d\d:\d\d:\d\d): (?P<msg>.*?)\n', re.DOTALL)
        with open(log_file, 'r') as log_f:
            df_log = pd.DataFrame(rg_log_msg.findall(log_f.read()),
                                  columns=['tipo', 'func', 'ts', 'msg'])
        df_log.drop('func', axis=1, inplace=True)
        df_log['tipo'] = df_log['tipo'].astype('category')
        df_log['ts'] = df_log['ts'].apply(lambda x: dt.datetime.strptime(x, '%d/%m/%Y %H:%M:%S'))
        df_log.loc[df_log.msg.str.startswith('Tªs --> '), 'temp'] = True
        df_log.loc[df_log.msg.str.startswith('SENDED: '), 'debug_send'] = True
        b_warn = df_log['tipo'] == 'WARNING'
        df_log.loc[b_warn, 'no_red'] = df_log[b_warn].msg.str.startswith('OSError: [Errno 101]; C_UNREACHABLE:')
        # df_log.loc[b_warn, 'no_red'] = df_log[b_warn].msg.str.startswith('OSError:  La red es inaccesible')
        df_log['exec'] = df_log['msg'].str.contains(INIT_LOG_MARK).cumsum().astype(int)
        df_log = df_log.set_index('ts')
        if extract_temps:
            rg_temps = 'Tªs --> (?P<CPU>\d{1,2}\.\d) / (?P<GPU>\d{1,2}\.\d) ºC'
            df_log = df_log.join(df_log[df_log['temp'].notnull()].msg.str.extract(rg_temps, expand=True).astype(float))
        if verbose:
            clasific = df_log.groupby(['exec', 'tipo']).count().dropna(how='all').astype(int)
            log(clasific, 'ok', True, False)
            conteo_tipos = df_log.groupby('tipo').count()
            if 'ERROR' in conteo_tipos.index:
                log(df_log[df_log.tipo == 'ERROR'].dropna(how='all', axis=1), 'error', True, False)
            if 'INFO' in conteo_tipos.index:
                log(df_log[df_log.tipo == 'INFO'].dropna(how='all', axis=1), 'info', True, False)
        return df_log
    else:
        log("extract_log_file: '{}' doesn't exists".format(log_file), 'error', verbose, False)
        return pd.DataFrame([], columns=['tipo', 'func', 'ts', 'msg'])



# -*- coding: utf-8 -*-
from flask_wtf import Form
from wtforms import StringField
from wtforms.validators import DataRequired


class DummyForm(Form):
    """
    For post submits without any field...
    """
    pass


class FileForm(Form):
    """
    File explorer
    """
    pathfile = StringField('pathfile', validators=[DataRequired()])



# -*- coding: utf-8 -*-
import argparse
import os
import sys
from enerpi.base import BASE_PATH, log, NGINX_CONFIG_FILE, UWSGI_CONFIG_FILE
from enerpiweb import basedir, CONFIG, DATA_PATH, check_resource_files, app as application
from enerpi.prettyprinting import print_secc, print_cyan, print_red, print_magenta


FLASK_WEBSERVER_PORT = CONFIG.getint('ENERPI_WEBSERVER', 'FLASK_WEBSERVER_PORT', fallback=7777)
PERIOD_MINUTES_RSC_GEN = CONFIG.getint('ENERPI_WEBSERVER', 'RSC_GEN_EVERY_MINUTES', fallback=15)
USER_SERVER = CONFIG.get('ENERPI_WEBSERVER', 'USER_SERVER', fallback='www-data')


def make_cron_command_task_periodic_rscgen():
    """
    CRON task for generate web resources with enerpiplot.mule_rscgen.py
    Example command:
    */15 * * * * sudo -u www-data /home/pi/PYTHON/py35/bin/python
        /home/pi/PYTHON/py35/lib/python3.5/site-packages/enerpiweb/mule_rscgen.py -o

    :return: :str: cron_command
    """
    # cmd_server = '*/{period} * * * * ...'
    cmd_server = 'sudo -u {user_server} {python_pathbin}/python3 {path_enerpiplot}/mule_rscgen.py -o'
    local_params = dict(path_enerpiplot=os.path.abspath(os.path.join(BASE_PATH, '..', 'enerpiplot')),
                        period=PERIOD_MINUTES_RSC_GEN, user_server=USER_SERVER,
                        python_pathbin=os.path.dirname(sys.executable))
    return cmd_server.format(**local_params)


def _make_webserver_config(overwrite=False):
    """
    Genera y escribe en DATA_PATH la configuración de la aplicación web para NGINX y UWSGI-EMPEROR.
    Muestra los comandos para realizar los enlaces correspondientes en /etc/nginx y  /etc/uwsgi-emperor.
    """
    path_config_uwsgi = os.path.join(DATA_PATH, UWSGI_CONFIG_FILE)
    path_config_nginx = os.path.join(DATA_PATH, NGINX_CONFIG_FILE)

    if overwrite or not (os.path.exists(path_config_nginx) and os.path.exists(path_config_uwsgi)):
        nginx_template = os.path.join(basedir, 'templates', 'nginx_conf_mask.txt')
        uwsgi_template = os.path.join(basedir, 'templates', 'uwsgi_ini_mask.txt')
        with open(nginx_template, 'r') as f:
            nginx_raw = f.read()
        with open(uwsgi_template, 'r') as f:
            uwsgi_raw = f.read()
        local_params = dict(file_location=DATA_PATH,
                            filename=UWSGI_CONFIG_FILE,
                            path_enerpiplot=os.path.abspath(os.path.join(basedir, '..', 'enerpiplot')),
                            path_enerpiweb=basedir,
                            path_venv=os.path.abspath(os.path.join(os.path.dirname(sys.executable), '..')))
        uwsgi_conf = uwsgi_raw.format(**local_params)
        nginx_conf = nginx_raw.replace('{file_location}', local_params['file_location']
                                       ).replace('{path_enerpiweb}', local_params['path_enerpiweb']
                                                 ).replace('{filename}', NGINX_CONFIG_FILE)

        # Copy config files to DATA_PATH
        check_resource_files(DATA_PATH, verbose=True)
        with open(os.path.join(DATA_PATH, UWSGI_CONFIG_FILE), 'w') as f:
            f.write(uwsgi_conf)
        with open(os.path.join(DATA_PATH, NGINX_CONFIG_FILE), 'w') as f:
            f.write(nginx_conf)

        # Show Info
        print_secc('NGINX Config generated:')
        print_cyan(nginx_conf)
        print_secc('UWSGI INI Config generated:')
        print_cyan(uwsgi_conf)
    else:
        # Show Info
        print_secc('NGINX Config at "{}":'.format(path_config_nginx))
        print_cyan(open(path_config_nginx, 'r').read())
        print_secc('UWSGI INI Config at "{}":'.format(path_config_uwsgi))
        print_cyan(open(path_config_uwsgi, 'r').read())
    print_red('\n* Append the NGINX config to your actual server, or make the next symlink:\n **"{}"**'
              .format('sudo ln -s {}/{} /etc/nginx/sites-enabled/'.format(DATA_PATH, NGINX_CONFIG_FILE)))
    print_red('* Make a symlink to deposit the UWSGI-Emperor configuration:\n **"{}"**\n'
              .format('sudo ln -s {}/{} /etc/uwsgi-emperor/vassals/'.format(DATA_PATH, UWSGI_CONFIG_FILE)))
    print_magenta('* To start the webserver, restart NGINX & UWSGI-EMPEROR:\n **"{}"**\n **"{}"**\n'
                  .format('sudo service nginx restart', 'sudo service uwsgi-emperor restart'))


def main():
    """
    CLI para ejecutar el webserver a mano vía flask en puerto 7777, o para mostrar/instalar la configuración
    de UWSGI y NGINX para servir la aplicación

    """
    p = argparse.ArgumentParser(description="\033[1m\033[5m\033[32m{}\033[0m\n\n".format('ENERPI Web Server'),
                                epilog='\033[34m\n*** By default, ENERPIWEB starts with flask server ***\n' +
                                       '\033[0m', formatter_class=argparse.RawTextHelpFormatter)
    p.add_argument('-p', '--port', action='store', type=int, metavar='P', default=FLASK_WEBSERVER_PORT,
                   help='✏  Flask PORT. Default: {}'.format(FLASK_WEBSERVER_PORT))
    p.add_argument('-d', '--debug', action='store_true', help='☕  DEBUG Mode')
    p.add_argument('-i', '--info', action='store_true', help='︎ℹ️  Show config params for NGINX + UWSGI')
    p.add_argument('--install', action='store_true',
                   help='⚒ Install CRON task for WEB RSC generator every {} minutes'.format(PERIOD_MINUTES_RSC_GEN))
    p.add_argument('--uninstall', action='store_true', help='⚒ Uninstall CRON task for WEB RSC generator')
    args = p.parse_args()

    if args.info:
        _make_webserver_config(overwrite=False)
    elif args.install or args.uninstall:
        from enerpi.config.crontasks import set_command_periodic, clear_cron_commands
        # INSTALL / UNINSTALL CRON TASKS & KEY
        cmd_server = make_cron_command_task_periodic_rscgen()
        if args.install:
            log('** (Re-)Create webserver config files:', 'ok', True, False)
            _make_webserver_config(overwrite=True)
            log('** Installing CRON task for web resources generation every {} minutes:\n"{}"'
                .format(PERIOD_MINUTES_RSC_GEN, cmd_server), 'ok', True, False)
            set_command_periodic(cmd_server, comment='Generador de recursos para ENERPIWEB',
                                 minute=PERIOD_MINUTES_RSC_GEN, verbose=True)
        else:
            log('** Deleting CRON task for web resources generation every X minutes:\n"{}"'
                .format(cmd_server), 'warn', True, False)
            clear_cron_commands([cmd_server], verbose=True)
            print_red('\n* To stop the webserver, remove config files from {}:\n **"{}"**\n **"{}"**\n'
                      .format('NGINX & UWSGI-EMPEROR',
                              'sudo rm /etc/uwsgi-emperor/vassals/{}'.format(UWSGI_CONFIG_FILE),
                              'sudo rm /etc/nginx/sites-enabled/{}'.format(NGINX_CONFIG_FILE)))
    else:
        log('EJECUTANDO FLASK WSGI A MANO en P:{}!'.format(FLASK_WEBSERVER_PORT), 'ok', True, False)
        application.run(host="0.0.0.0", port=args.port, processes=1, threaded=True, debug=args.debug)


if __name__ == '__main__':
    main()
